use log::{ debug, info, warn, error };
use dbus::{
    message::MatchRule,
    channel::MatchingReceiver
};
use dbus_tree::Factory;
use ifaces::*;
use tokio::sync::RwLock;
use std::sync::Arc;

mod ifaces;

const DBUS_NAME: &'static str = "org.ddnss.sfs.mc";
// Interface tokio runtime
static mut IFACE_RT: Option<Arc<tokio::runtime::Runtime>> = None;
static IFACE_RT_INIT: std::sync::Once = std::sync::Once::new();

fn iface_name(iface: &str) -> String {
    DBUS_NAME.to_owned() + "." + iface
}

// fn iface_rt() -> Arc<tokio::runtime::Runtime> {
fn iface_rt() -> tokio::runtime::Handle {
    IFACE_RT_INIT.call_once(|| {
        unsafe {
            // IFACE_RT = Some(Arc::new(tokio::runtime::Builder::new_multi_thread()
            IFACE_RT = Some(Arc::new(tokio::runtime::Builder::new()
                .thread_name("iface-runtime-worker")
                .build()
                .unwrap()
            ));
        }
    });

    unsafe {
        // IFACE_RT.clone().unwrap()
        IFACE_RT.clone().map(|x| x.handle().clone()).unwrap()
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    // env_logger initialization
    // If running with debug profile and `RUST_LOG` environment variable isn't set, set it to "debug"
    if cfg!(debug_assertions) && std::env::var_os(env_logger::DEFAULT_FILTER_ENV).is_none() {
        std::env::set_var(env_logger::DEFAULT_FILTER_ENV, "debug");
    }
    env_logger::init();

    debug!("PID: {}, User: {}", std::process::id(), std::env::var("USER").unwrap_or("N/A".into()));

    // Build interfaces
    let f = Factory::new_sync::<ServiceData>();
    let mut iface_map = std::collections::HashMap::new();
    iface_map.insert(iface_name("Service"), Arc::new(
        f.interface(iface_name("Service"), ())
            .add_m(f.method_sync("Start", (), handle_iface!(m, |data| {
                data.read().await
                    .start().await
                    .map(|_| vec![m.msg.method_return()])
                    .map_err(|err| err.into())
            })))
            .add_m(f.method_sync("Stop", (), handle_iface!(m, |data| {
                data.read().await
                    .stop().await
                    .map(|_| vec![m.msg.method_return()])
                    .map_err(|err| err.into())
            })))
            .add_m(f.method_sync("Restart", (), handle_iface!(m, |data| {
                data.read().await
                    .restart().await
                    .map(|_| vec![m.msg.method_return()])
                    .map_err(|err| err.into())
            })))
            .add_m(f.method_sync("IsRunning", (), handle_iface!(m, |data| {
                Ok(vec![m.msg.method_return().append1(
                    data.read().await
                        .is_running().await
                )])
            })).out_arg("b"))
    ));

    // Add interfaces to object paths in a new tree
    let tree = f.tree(())
        .add(f.object_path("/", Arc::new(RwLock::new(Service::Daemon)))
            .introspectable()
            .add(iface_map[&iface_name("Service")].clone())
        )
        .add(f.object_path("/minecraft", Arc::new(RwLock::new(Service::Minecraft)))
            .introspectable()
            .add(iface_map[&iface_name("Service")].clone())
        );

    // Connect with D-Bus
    let (resource, c) = dbus_tokio::connection::new_system_sync()?;
    debug!("D-Bus unique name: {}", c.unique_name());

    // Spawn a new task for the D-Bus resource
    tokio::spawn(async move {
        let err = resource.await;
        panic!("Lost connection to D-Bus: {}", err);
    });

    // Aquire name on D-Bus
    match c.request_name(DBUS_NAME, false, true, false).await {
        Ok(_) => info!("Listening (D-Bus) on {}", DBUS_NAME),
        Err(err) => {
            error!("Couldn't request name on D-Bus: {}", err);
            std::process::exit(1);
        }
    };

    // Receive method calls
    c.start_receive(MatchRule::new_method_call(), Box::new(move |msg, c| {
        use dbus::channel::Sender;
        if let Some(replies) = tree.handle(&msg) {
            for r in replies {
                let _ = c.send(r);
            }
        }
        true
    }));

    // Has been stabilized in Rust 1.48
    core::future::pending::<()>().await;
    unreachable!();
}